/* Performance counters for code profiling */


// (c) 1997 Sami Sallinen
// Editing: Marcel Hendrix, Dec 2000. 


#include "perfctrs.h"
#include <windows.h>
#include <winioctl.h>
#include "stdio.h"
#include "string.h"
#include "prfioctl.h"


// routine to get the CPU clock counter
unsigned __int64 TimeStamp (void) {
  unsigned highpart, lowpart;
  __asm {
    push EAX
    push EDX
    // the RDTSC instruction is not 
    // supported by VC++4.0 inline asm
    _asm _emit 0x0F
    _asm _emit 0x31
    mov lowpart, eax
    mov highpart, edx
    pop EDX
    pop EAX }
  return ((unsigned __int64)highpart << 32) | (unsigned __int64)lowpart; }

// read performance counter 0 or 1 [or 2 or 3] depending on ctr
unsigned __int64 ReadPerfCtr (int ctr) {
  unsigned highpart, lowpart;
  if (ctr > CTRMAX) ctr = 0; // make ctr 0 or 1 [or 2 or 3]
  __asm {
    push EAX
    push EDX
    push ECX
    mov  ECX, ctr
    // the RDPMC instruction is not 
    // supported by VC++4.0 inline asm
    _asm _emit 0x0F
    _asm _emit 0x33
    mov lowpart, eax
    mov highpart, edx    
    pop ECX
    pop EDX
    pop EAX }
  return ((unsigned __int64)highpart << 32) | lowpart; };

// ------- Exported function
struct CPU_ID *GetCPU_ID () {
  static struct CPU_ID ret;
  unsigned long *lp;
  unsigned long c_cpuid = 0;
  unsigned long c_eax;
  unsigned long c_ebx;
  unsigned long c_ecx;
  unsigned long c_edx;
  __asm {
    pusha;
    mov eax, c_cpuid;
    // the CPUID instruction is not 
    // supported by VC++4.0 inline asm
    __asm _emit 0x0F
    __asm _emit 0xA2
    mov c_eax, eax
    mov c_ebx, ebx
    mov c_ecx, ecx
    mov c_edx, edx
    popa; 
  }
  ret.maxIdCall = c_eax;
  ret.vendorString[12] = 0;
  lp = (unsigned long * )&ret.vendorString[0];
  *lp++ = c_ebx;
  *lp++ = c_edx;
  *lp++ = c_ecx;
  if (ret.maxIdCall > 0) {
    c_cpuid = 1;
    __asm {
      pusha;
      mov eax, c_cpuid;
      __asm _emit 0x0F
      __asm _emit 0xA2
      mov c_eax, eax
      mov c_ebx, ebx
      mov c_ecx, ecx
      mov c_edx, edx
      popa; 
	}
    ret.stepping =  c_eax & 0x000F;
    ret.model    = (c_eax & 0x00F0) >> 4;
    ret.family   = (c_eax & 0x0F00) >> 8;
    ret.type     = (c_eax & 0x3000) >> 12;
    ret.features =  c_edx; } 
  else {
    ret.stepping = 0;
    ret.model = 0;
    ret.family = 0;
    ret.type = 0;
    ret.features = 0; }
  return &ret; }

// ------- Exported function
const char *PrintID (struct CPU_ID *id) {
  static char buf[10240];
  sprintf(buf, "VendorID: %s, MaxCPUID %d, type %d, family %d, model %d, stepping %d\n",
          id->vendorString, id->maxIdCall, id->type, id->family, id->model, id->stepping);
#define SP(a, b) if (id->features & a) strcat(buf, b);
  SP(FE_FPU, "FPU on-chip\n");
  SP(FE_VME, "Virtual-8086 mode enhancement\n");
  SP(FE_DE,  "Debugging extensions\n");
  SP(FE_PSE, "Page size extensions\n");
  SP(FE_TSC, "Time-stamp-counter\n");
  SP(FE_MSR, "Model Specific Register support\n");
  SP(FE_PAE, "Physical Address Extensions\n");
  SP(FE_MCE, "Machine Check Exception\n");
  SP(FE_CXS, "CMPXCHG8B\n");
  SP(FE_APIC,"APIC on-chip\n");
  SP(FE_MTRR,"Memory type range register\n");
  SP(FE_PGE, "PTE Global bit\n");
  SP(FE_MCA, "Machine check architecture\n");
  SP(FE_CMOV,"CMOV instruction\n");
  SP(FE_MMX, "MMX(tm) Technology\n");
#undef SP
  return buf; }

// ------- Exported function
bool CanDoTimeStamp () {
  struct CPU_ID *id = GetCPU_ID();
  if (id->features & FE_TSC) return true;
  return false; }

// ------- Exported function
bool CanDoPerfMon () {
  struct CPU_ID *id = GetCPU_ID();
  if (id->features & FE_MSR) return true;
  return false; }

/* ------------------------------------------------------------

      Reading and writing the model-specific registers

------------------------------------------------------------- */

// ------- Global state
HANDLE m_commHandle;
bool bInitOk;
unsigned long m_prevCR4;

// ------- Local function
void MSR_IO_EnablePMCandTSC() {
  ULONG inputData = 0x100;	// clear TSD bit from CR4, set PCE bit  
  ULONG outputData;
  ULONG nRead;
  ULONG ioctlCode;
  if (!bInitOk || !CanDoTimeStamp() || !CanDoPerfMon()) return; 
  ioctlCode = (ULONG)IOCTL_ENABLE_CR4_BITS;
  if (!DeviceIoControl(
			m_commHandle,			// Handle to device
			ioctlCode,				// IO Control code for Read
			&inputData,				// Buffer to driver.
			sizeof(inputData),		// Length of buffer in bytes.
			&outputData,			// Buffer from driver.
			sizeof(outputData),		// Length of buffer in bytes.
			&nRead,					// Bytes placed in DataBuffer.
			NULL)) {int fail = 1;};	// NULL means wait until operation completes.
  m_prevCR4 = outputData; }

// ------- Local function
void MSR_IO_DisablePMCandTSC() {
  ULONG inputData = 0; // clear TSD bit from CR4, set PCE bit  
  ULONG outputData;
  ULONG nRead;
  ULONG ioctlCode;
  if (!bInitOk || !CanDoTimeStamp() || !CanDoPerfMon()) return;
  ioctlCode = (ULONG)IOCTL_ENABLE_CR4_BITS;
  if (!DeviceIoControl(
			m_commHandle,			// Handle to device
            ioctlCode,				// IO Control code for Read
            &inputData,				// Buffer to driver.
            sizeof(inputData),		// Length of buffer in bytes.
            &outputData,			// Buffer from driver.
            sizeof(outputData),		// Length of buffer in bytes.
            &nRead,					// Bytes placed in DataBuffer.
            NULL)) {int fail = 1;};	// NULL means wait until operation completes.  
  m_prevCR4 = outputData; }

// ------- Exported function
int InitMSR (void) {  
  HANDLE h = GetCurrentProcess();
  m_commHandle = CreateFile(
		"\\\\.\\PerfMon",	// Open the Device
		GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
  if (m_commHandle == INVALID_HANDLE_VALUE) { bInitOk = false; CloseHandle(m_commHandle); return -1; }
  bInitOk = true;  
  MSR_IO_EnablePMCandTSC(); 
  return IAM; }

// ------- Exported function
// This should set the CR4 bit as it was found, but it doesn't.
void ExitMSR (void) {
  if (bInitOk) { 
	  MSR_IO_DisablePMCandTSC();
      CloseHandle(m_commHandle);}}

// ------- Exported function
__int64 ReadMSR(unsigned long reg) {
  ULONG ioctlCode;
  ULONG inputData[1];
  ULONG outputData[2];
  ULONG nRead;
  inputData[0] = reg;
  if (!bInitOk) return -1;
  ioctlCode = (ULONG)IOCTL_READ_MSR;
  if  (!DeviceIoControl(
			m_commHandle,			// Handle to device
			ioctlCode,				// IO Control code for Read
			inputData,				// Buffer to driver.
			sizeof(inputData),		// Length of buffer in bytes.
			outputData,				// Buffer from driver.
			sizeof(outputData),		// Length of buffer in bytes.
			&nRead,					// Bytes placed in DataBuffer.
			NULL)) { return -1; }	// NULL means wait until operation completes.
  if (nRead != 8) return -1;
  return outputData[0] | ((__int64)(outputData[1]) << 32); }

// ------- Exported function
void WriteMSR(unsigned long reg, __int64 data) {
  ULONG inputData[3];
  ULONG outputData[2];
  ULONG nRead;
  ULONG ioctlCode;
  inputData[0] = reg;
  inputData[1] = (unsigned long)(data & 0xFFFFFFFF);
  inputData[2] = (unsigned long)(data >> 32);
  if (!bInitOk) return;  
  ioctlCode = (ULONG)IOCTL_WRITE_MSR;
  if  (!DeviceIoControl(
			m_commHandle,			// Handle to device
			ioctlCode,				// IO Control code for Read
            inputData,				// Buffer to driver.
            sizeof(inputData),		// Length of buffer in bytes.
            outputData,				// Buffer from driver.
            sizeof(outputData),		// Length of buffer in bytes.
            &nRead,					// Bytes placed in DataBuffer.
            NULL)) {int fail = 1;}}	// NULL means wait until operation completes.  
  
/*
 Values to be written to MSR_EVNTSEL0 and MSR_EVNTSEL1
 We do currently not allow to set the PC, Inv and CMASK fields

 Bits    Name            Description
 0:7     Event Select    number of event to be monitored
 8:15    Unit mask       MOESI specification, 0 is OK for most EventSelectors.
 16      User            count events in user mode   (privilege levels 1-3)
 17      OS              count events in kernel mode (privilege level 0)
 18      E               Edge detect when 1. Count occurrences when 0.
 19      PC              Enable signalling of counter overflows
 20      INT             Enable interrupt on counter overflow
 22      Enable          Enable counting (differs from Intel, Athlon enables each counter individually)
 23      Inv             1 inverts counter mask condition
 24:31   CMASK           Counter mask. Sets threshold for multiple events/clock (?). Zero works OK.
*/

// ------- Exported function
void SetEventCounter(int ctr, enum Event event, bool user, bool os, bool numbers) {
  unsigned long msr; __int64 val = (unsigned long) event;
  if (ctr > CTRMAX) ctr = 0;
  switch (ctr) {
	  case 0: msr = MSR_EVNTSEL0; break;
	  case 1: msr = MSR_EVNTSEL1; break;
	  case 2: msr = MSR_EVNTSEL2; break;
	  case 3: msr = MSR_EVNTSEL3; break; }
  if (user)		val |= 1 << 16;
  if (os)		val |= 1 << 17;
  if (!numbers) val |= 1 << 18;
  WriteMSR(msr, val |= 1 << 22); }

// ------- Exported function
void DisableEventCounting() {
  __int64 val = ReadMSR(MSR_EVNTSEL0);
  WriteMSR(MSR_EVNTSEL0, val &= ~(1 << 22));}

/* End of File */