/* Performance counters for code profiling */


// (c) 1997 Sami Sallinen
// Editing: Marcel Hendrix, Dec 2000.


typedef unsigned int bool;
#define true  1
#define false 0

// I've ONLY tested this chip!
#define AMD_ATHLON

// model-specific registers on Pentia
#define MSR_TSC 0x10
#define MSR_APICBASE 0x1B

// IAM : byte1 == CPU, byte0 == #counters

#ifdef AMD_ATHLON
#  define MSR_EVNTSEL0	0xC0010000
#  define MSR_EVNTSEL1	0xC0010001
#  define MSR_EVNTSEL2	0xC0010002
#  define MSR_EVNTSEL3	0xC0010003
#  define MSR_PERFCTR0	0xC0010004
#  define MSR_PERFCTR1	0xC0010005
#  define MSR_PERFCTR2	0xC0010006
#  define MSR_PERFCTR3	0xC0010007
#  define CTRMAX		3
static int IAM = 0x74;
#endif

enum feature_flag {
    FE_FPU  = 0x000001,		// FPU ON_CHIP
    FE_VME  = 0x000002,		// Virtual-8086 mode enhancement
    FE_DE   = 0x000004,		// Debugging extensions
    FE_PSE  = 0x000008,		// Page size extensions
    FE_TSC  = 0x000010,		// Time-stamp-counter
    FE_MSR  = 0x000020,		// Model Specific Register support
    FE_PAE  = 0x000040,		// Physical Address Extensions
    FE_MCE  = 0x000080,		// Machine Check Exception
    FE_CXS  = 0x000100,		// CMPXCHG8B instruction
    FE_APIC = 0x000200,		// APIC on-chip
    FE_MTRR = 0x001000,		// Mem. type range register
    FE_PGE  = 0x002000,		// PTE Global bit
    FE_MCA  = 0x004000,		// Machine check architecture
    FE_CMOV = 0x008000,		// CMOV instruction
    FE_MMX  = 0x800000 };	// MMX(tm) Technology

enum MemMask {
 MODIFIED	= 0x1000,
 OWNER		= 0x0800,
 EXCLUSIVE	= 0x0400,
 SHARED		= 0x0200,
 INVALID	= 0x0100 
};

enum Event {
 // Pentium Pro countable events
 // Flagged with [??] : not defined on Athlon. Some seem to work nevertheless.
 EVT_DATA_CACHE_ACCESS              = 0x0040,
 EVT_DATA_CACHE_MISSES              = 0x0041,
 
 EVT_DATA_CACHE_REFILL_L2			= 0x1F42, // all	
 EVT_DATA_CACHE_REFILL_L2_M			= 0x0042 | MODIFIED,
 EVT_DATA_CACHE_REFILL_L2_O			= 0x0042 | OWNER,
 EVT_DATA_CACHE_REFILL_L2_E			= 0x0042 | EXCLUSIVE,
 EVT_DATA_CACHE_REFILL_L2_S			= 0x0042 | SHARED,
 EVT_DATA_CACHE_REFILL_L2_I			= 0x0042 | INVALID,

 EVT_DATA_MEM_REFS                  = 0x1F43, // all
 EVT_DATA_MEM_REFS_M				= 0x0043 | MODIFIED,
 EVT_DATA_MEM_REFS_O				= 0x0043 | OWNER,
 EVT_DATA_MEM_REFS_E				= 0x0043 | EXCLUSIVE,
 EVT_DATA_MEM_REFS_S				= 0x0043 | SHARED,
 EVT_DATA_MEM_REFS_I				= 0x0043 | INVALID,

 EVT_DATA_CACHE_WRITEBACKS          = 0x1F44, // all
 EVT_DATA_CACHE_WRITEBACKS_M		= 0x0044 | MODIFIED,
 EVT_DATA_CACHE_WRITEBACKS_O		= 0x0044 | OWNER,
 EVT_DATA_CACHE_WRITEBACKS_E		= 0x0044 | EXCLUSIVE,
 EVT_DATA_CACHE_WRITEBACKS_S		= 0x0044 | SHARED,
 EVT_DATA_CACHE_WRITEBACKS_I		= 0x0044 | INVALID,

 EVT_L1_DLTB_misses_L2_DLTB_hits	= 0x0045,
 EVT_L1_AND_L2_DTLB_misses			= 0x0046,
 EVT_MISALIGNED_DATA_REFERENCES		= 0x0047,

 EVT_IFU_IFETCH						= 0x0080, // Number of instruction fetches, both cacheable and non cacheable.
 EVT_IFU_IFETCH_MISS				= 0x0081, // Number of instruction fetch misses.
 EVT_L1_ITLB_MISSES					= 0x0084,
 EVT_ITLB_MISS						= 0x0085, // Number of ITLB misses.

 EVT_INST_RETIRED                   = 0x00C0, // Number of instructions retired.
 EVT_UOPS_RETIRED                   = 0x00C1, // Number of UOPs retired.
 EVT_BR_INST_RETIRED                = 0x00C2, // Number of branch instructions retired.

 EVT_RETIRED_MISPREDICTED_BRANCHES	= 0x00C3,
 EVT_RETIRED_TAKEN_BRANCH_MISPRED	= 0x00C4, // Number of mispredicted branches retired.
 EVT_RETIRED_FAR_CONTROL_TRANSFERS	= 0x00C5,
 EVT_RETIRED_RESYNC_BRANCHES		= 0x00C7,

 EVT_INTS_MASKED					= 0x00CD,
 EVT_INTS_MASKED_WHILE_PENDING		= 0x00CE,
 EVT_INTS_TAKEN						= 0x00CF
};

struct CPU_ID {
  char vendorString[13];
  int maxIdCall;
  int stepping;
  int model;
  int family;
  int type;
  int features; };

// Inquire and Return the cpu capabilities in a CPU_ID struct
// Return true if TimeStamp() can be used (result based on CPUID)
// Return true if performance monitoring counters can be used (result based on CPUID)
__declspec (dllexport) struct CPU_ID *GetCPU_ID (void);
__declspec (dllexport) const char *PrintID (struct CPU_ID *id);
__declspec (dllexport) bool CanDoTimeStamp (void);
__declspec (dllexport) bool CanDoPerfMon (void);

// Sets event counter ctr to count the specified event
//  user == true -> count events in user mode
//  os == true -> count events in kernel mode
//  numbers == true -> count event occurrences, not event duration
__declspec (dllexport) void SetEventCounter (int ctr, enum Event event, bool user, bool os, bool numbers);
__declspec (dllexport) void DisableEventCounting (void);

// Read and write model-specific registers
// Warning: a non-existent msr will blue-screen your computer
__declspec (dllexport) __int64 ReadMSR (unsigned long reg);
__declspec (dllexport) void WriteMSR (unsigned long reg, __int64 data);
__declspec (dllexport) int InitMSR (void);  
__declspec (dllexport) void ExitMSR (void);
__declspec (dllexport) unsigned __int64 TimeStamp (void);
__declspec (dllexport) unsigned __int64 ReadPerfCtr (int);


/* End of File */